package ru.dbotthepony.mc.otm.android

import net.minecraft.nbt.CompoundTag
import net.minecraftforge.common.util.INBTSerializable
import net.minecraftforge.event.entity.living.LivingHurtEvent
import ru.dbotthepony.mc.otm.capability.MatteryPlayerCapability
import ru.dbotthepony.mc.otm.core.readNbt
import ru.dbotthepony.mc.otm.container.set
import ru.dbotthepony.mc.otm.core.set
import ru.dbotthepony.mc.otm.core.writeNbt
import java.io.InputStream
import java.io.OutputStream

abstract class AndroidFeature(val type: AndroidFeatureType<*>, val android: MatteryPlayerCapability) : INBTSerializable<CompoundTag> {
	val entity get() = android.ply

	/**
	 * Whenever there are changes to network
	 */
	var isDirty = false

	open var level = 0
		set(value) {
			if (value != field) {
				field = value
				applyModifiers()
			}
		}

	open fun tickClient() {}
	open fun tickServer() {}

	/**
	 * Called when it is required to network everything again
	 */
	open fun invalidateNetwork() {}

	open fun applyModifiers() {}
	open fun removeModifiers() {}

	open fun onHurt(event: LivingHurtEvent) {}

	open fun writeNetwork(stream: OutputStream) {
		stream.writeNbt(serializeNBT())
	}

	open fun readNetwork(stream: InputStream) {
		deserializeNBT(stream.readNbt())
	}

	override fun serializeNBT(): CompoundTag {
		return CompoundTag().also {
			it["level"] = level
		}
	}

	override fun deserializeNBT(nbt: CompoundTag) {
		level = nbt.getInt("level")
	}
}

class DummyAndroidFeature(type: AndroidFeatureType<*>, android: MatteryPlayerCapability) : AndroidFeature(type, android) {
	override fun serializeNBT(): CompoundTag {
		return CompoundTag()
	}

	override fun deserializeNBT(nbt: CompoundTag) {

	}
}

