package ru.dbotthepony.mc.otm.capability

import it.unimi.dsi.fastutil.objects.Object2IntAVLTreeMap
import it.unimi.dsi.fastutil.objects.ObjectArraySet
import net.minecraft.ChatFormatting
import net.minecraft.core.Direction
import net.minecraft.nbt.CompoundTag
import net.minecraft.nbt.ListTag
import net.minecraft.nbt.StringTag
import net.minecraft.network.chat.Component
import net.minecraft.resources.ResourceLocation
import net.minecraft.server.level.ServerPlayer
import net.minecraft.world.effect.MobEffect
import net.minecraft.world.effect.MobEffects
import net.minecraft.world.entity.Entity
import net.minecraft.world.entity.LivingEntity
import net.minecraft.world.entity.item.ItemEntity
import net.minecraft.world.entity.player.Player
import net.minecraft.world.item.ItemStack
import net.minecraft.world.item.enchantment.EnchantmentHelper.hasVanishingCurse
import net.minecraft.world.level.GameRules
import net.minecraftforge.common.capabilities.Capability
import net.minecraftforge.common.capabilities.ICapabilityProvider
import net.minecraftforge.common.util.INBTSerializable
import net.minecraftforge.common.util.LazyOptional
import net.minecraftforge.event.AttachCapabilitiesEvent
import net.minecraftforge.event.entity.living.LivingDeathEvent
import net.minecraftforge.event.entity.living.LivingDropsEvent
import net.minecraftforge.event.entity.living.LivingEvent.LivingTickEvent
import net.minecraftforge.event.entity.living.LivingHurtEvent
import net.minecraftforge.event.entity.player.EntityItemPickupEvent
import net.minecraftforge.event.entity.player.PlayerEvent
import net.minecraftforge.fml.event.lifecycle.FMLCommonSetupEvent
import org.apache.commons.lang3.mutable.MutableInt
import org.apache.logging.log4j.LogManager
import ru.dbotthepony.mc.otm.*
import ru.dbotthepony.mc.otm.android.AndroidFeature
import ru.dbotthepony.mc.otm.android.AndroidFeatureType
import ru.dbotthepony.mc.otm.android.AndroidResearch
import ru.dbotthepony.mc.otm.android.AndroidResearchType
import ru.dbotthepony.mc.otm.container.MatteryContainer
import ru.dbotthepony.mc.otm.core.*
import ru.dbotthepony.mc.otm.menu.ExoSuitInventoryMenu
import ru.dbotthepony.mc.otm.network.*
import ru.dbotthepony.mc.otm.registry.AndroidFeatures
import ru.dbotthepony.mc.otm.registry.MRegistry
import java.util.*
import kotlin.collections.ArrayDeque
import kotlin.collections.ArrayList

@Suppress("unused")
class MatteryPlayerCapability(val ply: Player) : ICapabilityProvider, INBTSerializable<CompoundTag> {
	private inner class PlayerMatteryContainer(size: Int) : MatteryContainer(size) {
		override fun setChanged(slot: Int, new: ItemStack, old: ItemStack) {
			super.setChanged(slot, new, old)
			// TODO: Minecraft has hard-coded inventory triggers to Inventory class
			// when nothing use Inventory class directly.

			// Why?
			// CriteriaTriggers.INVENTORY_CHANGED.trigger(this, this.getInventory(), p_143468_)
		}
	}

	val synchronizer = FieldSynchronizer()

	var hasExoSuit by synchronizer.bool()

	private val exoSuitSlotCountModifiersMap: MutableMap<UUID, Int> by synchronizer.Map(
		keyCodec = UUIDValueCodec,
		valueCodec = IntValueCodec,
		backingMap = Object2IntAVLTreeMap(),
		callback = {
			this.exoSuitSlotCountModifiers.recompute()
		}
	)

	val exoSuitSlotCountModifiers = UUIDIntModifiersMap(observer = observer@{
		if (ply !is ServerPlayer)
			return@observer

		if (it < 0) {
			exoSuitSlotCount = 0
		} else {
			exoSuitSlotCount = it
		}
	}, backingMap = this.exoSuitSlotCountModifiersMap)

	var exoSuitSlotCount by synchronizer.int(setter = setter@{ value, access, _ ->
		require(value >= 0) { "Invalid slot count $value" }

		if (value != access.read()) {
			access.write(value)
			_exoSuitMenu = null
			exoSuitContainer = PlayerMatteryContainer(value)
		}
	})

	var exoSuitContainer: MatteryContainer = PlayerMatteryContainer(0)
		private set(value) {
			_exoSuitMenu = null

			if (ply.containerMenu.slots.any { it.container == field }) {
				ply.closeContainer()
			}

			for (i in 0 until value.containerSize.coerceAtMost(field.containerSize)) {
				if (!field[i].isEmpty) {
					value[i] = field[i]
				}
			}

			for (i in value.containerSize until field.containerSize) {
				ply.spawnAtLocation(field[i])
			}

			field = value
		}

	fun dropExoSuitInventory(): List<ItemEntity> {
		val result = ArrayList<ItemEntity>()
		val oldCaptureDrops = ply.captureDrops(result) as MutableCollection<ItemEntity>?

		val iterator = exoSuitContainer.iterator().nonEmpty()

		for (item in iterator) {
			if (!hasVanishingCurse(item)) {
				ply.drop(item, true, false)
			}

			iterator.remove()
		}

		ply.captureDrops(oldCaptureDrops)
		oldCaptureDrops?.addAll(result)

		return result
	}

	var isExoSuitCraftingUpgraded by synchronizer.bool(setter = setter@{ value, access, _ ->
		if (value != access.read()) {
			access.write(value)
			_exoSuitMenu = null
		}
	})

	private var _exoSuitMenu: ExoSuitInventoryMenu? = null
		set(value) {
			if (field == ply.containerMenu) { ply.closeContainer() }
			field = value
		}

	val exoSuitMenu: ExoSuitInventoryMenu
		get() {
			if (_exoSuitMenu == null) {
				_exoSuitMenu = ExoSuitInventoryMenu(this)
			}

			return _exoSuitMenu!!
		}

	private var shouldSendIteration = false
	var iteration = 0
		private set

	private var lastDeathTick = 0

	private val deathLog = ArrayDeque<Pair<Int, Component>>()

	private val features = IdentityHashMap<AndroidFeatureType<*>, AndroidFeature>()
	private val networkQueue = ArrayList<Any>()
	private val queuedTicks = ArrayList<Runnable>()
	private var tickedOnce = false

	private var shouldPlaySound = false
	private val research = IdentityHashMap<AndroidResearchType<*>, AndroidResearch>()

	private var invalidateNetworkIn = 10

	var sleepTicks = 0
	val isEverAndroid: Boolean get() = isAndroid || willBecomeAndroid
	var lastJumpTicks = 14

	var willBecomeAndroid by synchronizer.bool()
	var isAndroid by synchronizer.bool()

	val androidEnergy = SynchronizedPowerWithBattery(ply, synchronizer, DEFAULT_MAX_ANDROID_POWER, DEFAULT_MAX_ANDROID_POWER)

	fun invalidateNetworkState() {
		invalidateNetworkIn = 10
	}

	fun becomeAndroidSoft() {
		if (isAndroid || willBecomeAndroid) return
		willBecomeAndroid = true
		(ply as? ServerPlayer)?.displayClientMessage(TranslatableComponent("otm.pill.message").withStyle(ChatFormatting.GRAY), false)
	}

	fun becomeAndroid() {
		if (isAndroid) return

		isAndroid = true
		willBecomeAndroid = false
		shouldPlaySound = false
		iteration = 0
		deathLog.clear()
		androidEnergy.batteryLevel = DEFAULT_MAX_ANDROID_POWER
		androidEnergy.maxBatteryLevel = DEFAULT_MAX_ANDROID_POWER
	}

	fun becomeAndroidAndKill() {
		if (isAndroid) return

		becomeAndroid()
		ply.hurt(MRegistry.DAMAGE_BECOME_ANDROID, ply.maxHealth * 2)
	}

	fun becomeHumane() {
		if (willBecomeAndroid) willBecomeAndroid = false
		if (!isAndroid) return

		isAndroid = false
		shouldPlaySound = false
		iteration = 0
		deathLog.clear()
		androidEnergy.batteryLevel = ImpreciseFraction.ZERO
		androidEnergy.maxBatteryLevel = DEFAULT_MAX_ANDROID_POWER
		dropBattery()
	}

	fun becomeHumaneAndKill() {
		if (willBecomeAndroid) willBecomeAndroid = false
		if (!isAndroid) return

		becomeHumane()
		ply.hurt(MRegistry.DAMAGE_BECOME_HUMANE, ply.maxHealth * 2)
	}

	fun obliviate(refund: Boolean = true) {
		for (instance in research.values) {
			if (instance.isResearched) {
				instance.unResearch()

				if (refund) {
					instance.refund(simulate = false)
				}
			}
		}

		val copy = ArrayList<AndroidFeature>(features.values.size).also { it.addAll(features.values) }

		for (feature in copy) {
			removeFeature(feature.type)
		}

		iteration = 0
		deathLog.clear()
	}

	fun <T : AndroidResearch> getResearch(type: AndroidResearchType<T>): T {
		return research.computeIfAbsent(type) {
			return@computeIfAbsent type.factory(this)
		} as T
	}

	private fun addFeature(feature: AndroidFeature): Boolean {
		if (features.containsKey(feature.type)) return false
		features[feature.type] = feature

		if (!ply.level.isClientSide) {
			queuedTicks.add(feature::applyModifiers)
		}

		if (ply is ServerPlayer) {
			sendNetwork(AndroidFeatureSyncPacket(feature))
		}

		return true
	}

	@Suppress("unchecked_cast")
	fun <T : AndroidFeature> addFeature(feature: AndroidFeatureType<T>): T {
		val get = features[feature]
		if (get != null) return get as T

		val factory = feature.create(this)

		features[feature] = factory

		if (!ply.level.isClientSide) {
			queuedTicks.add(factory::applyModifiers)
		}

		if (ply is ServerPlayer) {
			sendNetwork(AndroidFeatureSyncPacket(factory))
		}

		return factory
	}

	fun removeFeature(feature: AndroidFeatureType<*>): Boolean {
		val removed = features.remove(feature)

		if (removed != null) {
			if (!ply.level.isClientSide) {
				queuedTicks.add {
					removed.removeModifiers()
				}
			}

			if (ply is ServerPlayer) {
				sendNetwork(AndroidFeatureRemovePacket(removed.type))
			}
		}

		return removed != null
	}

	fun hasFeature(feature: AndroidFeatureType<*>): Boolean {
		return features.containsKey(feature)
	}

	fun hasFeatureLevel(feature: AndroidFeatureType<*>, level: Int): Boolean {
		val get = features[feature] ?: return false
		return get.level >= level
	}

	@Suppress("unchecked_cast")
	fun <T : AndroidFeature> getFeature(feature: AndroidFeatureType<T>): T? {
		return features[feature] as T?
	}

	fun onHurt(event: LivingHurtEvent) {
		for (feature in features.values) {
			feature.onHurt(event)
		}
	}

	fun <T : AndroidFeature> computeIfAbsent(feature: AndroidFeatureType<T>): T {
		return getFeature(feature) ?: addFeature(feature)
	}

	fun <T : AndroidFeature> getFeatureO(feature: AndroidFeatureType<T>): Optional<T> {
		val get = getFeature(feature)
		return if (get != null) Optional.of(get) else Optional.empty()
	}

	fun <T : AndroidFeature> ifFeature(feature: AndroidFeatureType<T>, consumer: (T) -> Unit) {
		val get = getFeature(feature)

		if (get != null) {
			consumer(get)
		}
	}

	override fun serializeNBT(): CompoundTag {
		val tag = CompoundTag()

		// iteration
		tag["iteration"] = iteration
		tag["shouldSendIteration"] = shouldSendIteration

		tag["deathLog"] = ListTag().also {
			for ((ticks, component) in deathLog) {
				it.add(CompoundTag().also {
					it["ticks"] = ticks
					it["component"] = StringTag.valueOf(Component.Serializer.toJson(component))
				})
			}
		}

		// exosuit
		tag["hasExoSuit"] = hasExoSuit
		tag["exoSuitContainer"] = exoSuitContainer.serializeNBT()
		tag["isExoSuitCraftingUpgraded"] = isExoSuitCraftingUpgraded
		tag["exoSuitSlotCountModifiers"] = exoSuitSlotCountModifiers.serializeNBT()

		// android
		tag["androidEnergy"] = androidEnergy.serializeNBT()

		tag["isAndroid"] = isAndroid
		tag["willBecomeAndroid"] = willBecomeAndroid

		val featureList = ListTag()
		val researchList = ListTag()

		for (feature in features.values) {
			featureList.add(feature.serializeNBT().also {
				it["id"] = feature.type.registryName!!.toString()
			})
		}

		for ((type, instance) in research) {
			researchList.add(instance.serializeNBT().also {
				it["id"] = type.registryName!!.toString()
			})
		}

		tag["features"] = featureList
		tag["research"] = researchList

		return tag
	}

	override fun deserializeNBT(tag: CompoundTag) {
		// iterations
		iteration = tag.getInt("iteration")
		shouldSendIteration = tag.getBoolean("shouldSendIteration")

		deathLog.clear()

		for (value in tag.getCompoundList("deathLog")) {
			val component = Component.Serializer.fromJson(value.getString("component"))

			if (component != null) {
				deathLog.add(value.getInt("ticks") to component)
			}
		}

		// exosuit
		hasExoSuit = tag.getBoolean("hasExoSuit")
		tag.map("exoSuitSlotCountModifiers", exoSuitSlotCountModifiers::deserializeNBT)

		exoSuitContainer.deserializeNBT(tag["exoSuitContainer"])
		isExoSuitCraftingUpgraded = tag.getBoolean("isExoSuitCraftingUpgraded")

		// android
		isAndroid = tag.getBoolean("isAndroid")
		willBecomeAndroid = tag.getBoolean("willBecomeAndroid")

		tag.map("androidEnergy", androidEnergy::deserializeNBT)

		features.clear()
		research.clear()

		for (featureTag in tag.getCompoundList("features")) {
			val feature = MRegistry.ANDROID_FEATURES.getValue(ResourceLocation(featureTag.getString("id")))

			if (feature?.isApplicable(this) == true) {
				val instance = feature.create(this)

				instance.deserializeNBT(featureTag)
				addFeature(instance)

				if (!ply.level.isClientSide) {
					queuedTicks.add(instance::applyModifiers)
				}
			}
		}

		for (researchTag in tag.getCompoundList("research")) {
			val research = MRegistry.ANDROID_RESEARCH.getValue(ResourceLocation(researchTag.getString("id")))

			if (research != null) {
				val instance = research.factory(this)
				instance.deserializeNBT(researchTag)
				this.research[research] = instance
			}
		}
	}

	fun dropBattery() {
		if (androidEnergy.item.isEmpty) return
		ply.spawnAtLocation(androidEnergy.item)
		androidEnergy.item = ItemStack.EMPTY
	}

	private fun sendNetwork(packet: Any) {
		if (ply is ServerPlayer) {
			if (tickedOnce) {
				MatteryPlayerNetworkChannel.send(ply, packet)
			} else {
				networkQueue.add(packet)
			}
		}
	}

	private fun tickInventory() {
		if (hasExoSuit) {
			for ((i, stack) in exoSuitContainer.withIndex()) {
				if (!stack.isEmpty) {
					stack.inventoryTick(ply.level, ply, i + 41, false)
				}
			}
		}
	}

	fun tickClient() {
		queuedTicks.clear()

		if (!ply.isAlive) {
			return
		}

		if (willBecomeAndroid) {
			if (ply.isSleeping) {
				sleepTicks++
			} else {
				sleepTicks = 0
			}
		}

		if (isAndroid) {
			for (feature in features.values) {
				feature.tickClient()
			}
		}

		tickInventory()
	}

	fun tick() {
		if (!ply.isAlive) return

		if (willBecomeAndroid) {
			if (ply.isSleeping) {
				sleepTicks++

				if (sleepTicks > SLEEP_TICKS_LIMIT) {
					becomeAndroid()
					shouldPlaySound = true
					sleepTicks = 0

					(ply as? ServerPlayer)?.displayClientMessage(TranslatableComponent("otm.pill.message_finish").withStyle(ChatFormatting.DARK_RED), false)
				}
			} else {
				sleepTicks = 0
			}
		}

		if (isAndroid) {
			if (ply.airSupply < ply.maxAirSupply)
				ply.airSupply = ply.maxAirSupply

			for (effect in UNAFFECTED_EFFECTS)
				if (ply.hasEffect(effect))
					ply.removeEffect(effect)

			// TODO: Maybe passive drain?
			// extractEnergyInner(BigDecimal.valueOf(new Random().nextDouble()), false);
			if (ply.isSwimming && !hasFeature(AndroidFeatures.AIR_BAGS)) {
				ply.isSwimming = false
			}

			androidEnergy.tick()

			val stats = ply.foodData

			while (stats.foodLevel < 18 && androidEnergy.extractEnergyInner(ENERGY_FOR_HUNGER_POINT, true) >= ENERGY_FOR_HUNGER_POINT) {
				androidEnergy.extractEnergyInner(ENERGY_FOR_HUNGER_POINT, false)
				stats.foodLevel = stats.foodLevel + 1
			}

			// "block" quick regeneration
			// also cause power to generate while in peaceful
			while (stats.foodLevel > 18 && androidEnergy.receiveEnergyInner(ENERGY_FOR_HUNGER_POINT, true) >= ENERGY_FOR_HUNGER_POINT) {
				androidEnergy.receiveEnergyInner(ENERGY_FOR_HUNGER_POINT, false)
				stats.foodLevel = stats.foodLevel - 1
			}

			val foodLevel = stats.foodLevel.toFloat()

			if (stats.saturationLevel < foodLevel) {
				val extracted = androidEnergy.extractEnergyInner(ENERGY_FOR_HUNGER_POINT * (foodLevel - stats.saturationLevel), false)
				stats.setSaturation(stats.saturationLevel + (extracted / ENERGY_FOR_HUNGER_POINT).toFloat())
			}

			if (stats.exhaustionLevel > 0f) {
				val extracted = androidEnergy.extractEnergyInner(ENERGY_FOR_HUNGER_POINT * (stats.exhaustionLevel / 4f), false)
				stats.setExhaustion(stats.exhaustionLevel - (extracted / ENERGY_FOR_HUNGER_POINT).toFloat() * 4f)
			}

			for (feature in features.values) {
				feature.tickServer()
			}
		}

		for (runnable in queuedTicks) {
			runnable.run()
		}

		queuedTicks.clear()

		if (invalidateNetworkIn > 0 && --invalidateNetworkIn == 0) {
			synchronizer.invalidate()

			for (instance in research.values) {
				instance.isDirty = true
				instance.invalidateNetwork()
			}

			for (feature in features.values) {
				feature.isDirty = true
				feature.invalidateNetwork()
			}
		}

		tickedOnce = true

		val payload = synchronizer.collectNetworkPayload()

		if (payload != null) {
			MatteryPlayerNetworkChannel.send(ply, MatteryPlayerFieldPacket(payload))
		}

		if (networkQueue.size != 0) {
			for (packet in networkQueue) {
				MatteryPlayerNetworkChannel.send(ply, packet)
			}

			networkQueue.clear()
		}

		for (instance in research.values) {
			if (instance.isDirty) {
				instance.isDirty = false
				sendNetwork(AndroidResearchSyncPacket(instance))
			}
		}

		for (instance in features.values) {
			if (instance.isDirty) {
				instance.isDirty = false
				sendNetwork(AndroidFeatureSyncPacket(instance))
			}
		}

		if (shouldSendIteration) {
			sendNetwork(PlayerIterationPacket(iteration, deathLog))
			shouldSendIteration = false
		}

		if (hasExoSuit && ply.containerMenu == ply.inventoryMenu) {
			exoSuitMenu.broadcastChanges()
		}

		tickInventory()
	}

	private val resolver = LazyOptional.of { this }

	override fun <T> getCapability(cap: Capability<T>, side: Direction?): LazyOptional<T> {
		return if (cap == MatteryCapability.MATTERY_PLAYER) {
			resolver.cast()
		} else LazyOptional.empty()
	}

	@Suppress("unused")
	companion object {
		val UNAFFECTED_EFFECTS = ObjectArraySet<MobEffect>()

		fun registerEffects(event: FMLCommonSetupEvent) {
			UNAFFECTED_EFFECTS.add(MobEffects.CONDUIT_POWER)
			UNAFFECTED_EFFECTS.add(MobEffects.HEAL)
			// maybe it makes things go haywire idk
			// UNAFFECTED_EFFECTS.add(MobEffects.HARM);
			UNAFFECTED_EFFECTS.add(MobEffects.REGENERATION)
			UNAFFECTED_EFFECTS.add(MobEffects.WATER_BREATHING)
			UNAFFECTED_EFFECTS.add(MobEffects.POISON)
			// even skeletons can be withered
			// UNAFFECTED_EFFECTS.add(MobEffects.WITHER);
			UNAFFECTED_EFFECTS.add(MobEffects.HEALTH_BOOST)
			UNAFFECTED_EFFECTS.add(MobEffects.ABSORPTION)
			UNAFFECTED_EFFECTS.add(MobEffects.SATURATION)
			UNAFFECTED_EFFECTS.add(MobEffects.DOLPHINS_GRACE)
		}

		fun onLivingTick(event: LivingTickEvent) {
			val ent = event.entity

			if (ent.level.isClientSide) {
				ent.getCapability(MatteryCapability.MATTERY_PLAYER).ifPresentK {
					it.tickClient()
				}
			} else {
				ent.getCapability(MatteryCapability.MATTERY_PLAYER).ifPresentK {
					it.tick()
				}
			}
		}

		fun onHurtEvent(event: LivingHurtEvent) {
			if (event.isCanceled) {
				return
			}

			event.entity.getCapability(MatteryCapability.MATTERY_PLAYER).ifPresentK { it.onHurt(event) }
		}

		const val CAPABILITY_KEY = "otm_player"
		val CAPABILITY_LOCATION = ResourceLocation(OverdriveThatMatters.MOD_ID, CAPABILITY_KEY)

		fun onAttachCapabilityEvent(event: AttachCapabilitiesEvent<Entity>) {
			val ent = event.`object`

			if (ent is Player) {
				event.addCapability(CAPABILITY_LOCATION, MatteryPlayerCapability(ent))
			}
		}

		fun onPlayerChangeDimensionEvent(event: PlayerEvent.PlayerChangedDimensionEvent) {
			event.entity.getCapability(MatteryCapability.MATTERY_PLAYER)
				.ifPresentK { it.invalidateNetworkState() }
		}

		private val waitingToDie = ArrayList<Pair<LivingEntity, MatteryPlayerCapability>>()

		fun onLivingDrops(event: LivingDropsEvent) {
			val indexOf = waitingToDie.indexOfFirst { event.entity == it.first }

			if (indexOf == -1)
				return

			val drops = waitingToDie[indexOf].second.dropExoSuitInventory()

			if (event.entity.captureDrops() == null) {
				event.drops.addAll(drops)
			}

			waitingToDie.removeAt(indexOf)
		}

		fun onPlayerDeath(event: LivingDeathEvent) {
			val ply = event.entity as? Player ?: return
			val mattery = ply.matteryPlayer ?: return

			if (mattery.lastDeathTick != ply.tickCount) {
				mattery.lastDeathTick = ply.tickCount
			} else {
				return
			}

			if (mattery.hasExoSuit && !ply.level.gameRules.getBoolean(GameRules.RULE_KEEPINVENTORY)) {
				waitingToDie.add(ply to mattery)

				addPostServerTickerOnce {
					val index = waitingToDie.indexOfFirst { it.first == ply }

					if (index != -1) {
						waitingToDie.removeAt(index)
						mattery.dropExoSuitInventory().forEach { ply.level.addFreshEntity(it) }
						LOGGER.warn("Unable to 'properly' drop $ply's ExoSuit inventory contents")
						LOGGER.warn("I blame Forge.")
					}
				}
			}

			if (!mattery.isAndroid) {
				return
			}

			mattery.iteration++
			mattery.shouldSendIteration = true
			mattery.deathLog.addLast(ply.tickCount to ply.combatTracker.deathMessage)

			while (mattery.deathLog.size > 6) {
				mattery.deathLog.removeFirst()
			}
		}

		fun onPlayerCloneEvent(event: PlayerEvent.Clone) {
			val it = event.entity.matteryPlayer ?: return

			var resolver = event.original.getCapability(MatteryCapability.MATTERY_PLAYER)

			if (!resolver.isPresent || resolver.resolve().isEmpty) {
				event.original.reviveCaps()
				resolver = event.original.getCapability(MatteryCapability.MATTERY_PLAYER)
			}

			if (!resolver.isPresent || resolver.resolve().isEmpty) {
				event.original.invalidateCaps()
				return
			}

			val original = resolver.resolve().get()

			if (original.willBecomeAndroid && event.isWasDeath) {
				original.becomeAndroid()
				(event.entity as? ServerPlayer)?.displayClientMessage(TranslatableComponent("otm.pill.message_finish").withStyle(ChatFormatting.DARK_RED), false)
			}

			it.deserializeNBT(original.serializeNBT())
			it.invalidateNetworkState()
			event.original.invalidateCaps()
		}

		val ENERGY_FOR_HUNGER_POINT = ImpreciseFraction(1000)
		const val SLEEP_TICKS_LIMIT = 80

		private val itemPickupTicks = WeakHashMap<Player, WeakHashMap<ItemEntity, MutableInt>>()

		private val LOGGER = LogManager.getLogger()

		fun onPickupEvent(event: EntityItemPickupEvent) {
			if (event.item.owner != null && event.item.owner != event.entity.uuid && event.item.age < 200 || event.item.item.isEmpty) {
				return
			}

			val it = event.entity.matteryPlayer ?: return

			if (!it.hasExoSuit) {
				return
			}

			val ticks = itemPickupTicks.computeIfAbsent(event.entity) { WeakHashMap() }.computeIfAbsent(event.item) { MutableInt() }
			ticks.increment()

			if (ticks.value < 10) {
				return
			}

			val leftover = it.exoSuitContainer.addItem(event.item.item, false)
			event.item.item.count = leftover.count
		}

		val DEFAULT_MAX_ANDROID_POWER = ImpreciseFraction(60000)
	}
}
