package ru.dbotthepony.mc.otm.client.screen.panels

import com.mojang.blaze3d.systems.RenderSystem
import com.mojang.blaze3d.vertex.PoseStack
import net.minecraft.client.renderer.GameRenderer
import net.minecraft.network.chat.Component
import net.minecraft.world.item.ItemStack
import net.minecraftforge.client.extensions.common.IClientItemExtensions
import org.lwjgl.opengl.GL11
import ru.dbotthepony.mc.otm.client.minecraft
import ru.dbotthepony.mc.otm.client.render.*
import ru.dbotthepony.mc.otm.client.screen.MatteryScreen
import ru.dbotthepony.mc.otm.core.RGBAColor

abstract class AbstractSlotPanel @JvmOverloads constructor(
	screen: MatteryScreen<*>,
	parent: EditablePanel?,
	x: Float = 0f,
	y: Float = 0f,
	width: Float = SIZE,
	height: Float = SIZE,
	open val noItemIcon: SkinElement? = null
) : EditablePanel(
	screen, parent, x, y, width, height
) {
	protected open fun renderSlotBackground(stack: PoseStack, mouseX: Float, mouseY: Float, partialTick: Float) {
		SLOT_BACKGROUND.render(stack, width = width, height = height)
	}

	protected open fun renderRegular(stack: PoseStack, itemstack: ItemStack, count_override: String? = null) {
		RenderSystem.setShader { GameRenderer.getPositionTexShader() }

		var zHeight = 0.0

		if (!itemstack.isEmpty) {
			RenderSystem.enableDepthTest()

			val system_stack = RenderSystem.getModelViewStack()

			system_stack.pushPose()
			system_stack.translate((absoluteX + 1).toDouble(), (absoluteY + 1).toDouble(), 0.0)
			RenderSystem.applyModelViewMatrix()
			RenderSystem.depthFunc(GL11.GL_LESS)

			// Thanks Mojang
			// Very cool
			// (for int x, int y, which are then cast into doubles anyway)
			screen.itemRenderer.blitOffset = accumulatedDepth - 100 // force item to draw only 50 units "above" background

			screen.itemRenderer.renderAndDecorateItem(
				requireNotNull(minecraft.player) { "yo, dude, what the fuck" },
				itemstack,
				0,
				0,
				(absoluteX + absoluteY * 1000).toInt()
			)

			RenderSystem.depthFunc(GL11.GL_ALWAYS)
			screen.itemRenderer.renderGuiItemDecorations(screen.font, itemstack, 0, 0, count_override)
			screen.itemRenderer.blitOffset = 0f

			// too big accumulations can lead to Z near clipping issues
			accumulate3DHeight(101f)
			zHeight = 101.0
			system_stack.popPose()
			RenderSystem.applyModelViewMatrix()
		}

		if (isHovered) {
			stack.pushPose()
			stack.translate(0.0, 0.0, zHeight)
			drawColor = SLOT_HIGHLIGHT
			drawRect(stack, 1f, 1f, width - 1, height - 1)
			stack.popPose()
		}
	}

	protected abstract fun getItemStack(): ItemStack

	protected open fun getItemStackTooltip(stack: ItemStack): List<Component> {
		return screen.getTooltipFromItem(stack)
	}

	override fun innerRender(stack: PoseStack, mouseX: Float, mouseY: Float, partialTick: Float) {
		renderSlotBackground(stack, mouseX, mouseY, partialTick)
		val itemStack = getItemStack()
		renderRegular(stack, itemStack)

		if (itemStack.isEmpty) {
			noItemIcon?.render(stack, width = width, height = height)
		}
	}

	override fun innerRenderTooltips(stack: PoseStack, mouse_x: Float, mouse_y: Float, flag: Float): Boolean {
		if (isHovered) {
			val itemstack = getItemStack()

			if (!itemstack.isEmpty) {
				// val font = RenderProperties.get(itemstack).getFont(itemstack)
				val font = (itemstack.item.renderPropertiesInternal as? IClientItemExtensions)?.getFont(itemstack, IClientItemExtensions.FontContext.TOOLTIP)

				// TODO: WHERE????????????
				// GuiUtils.preItemToolTip(itemstack);
				screen.renderComponentTooltip(
					stack,
					getItemStackTooltip(itemstack),
					mouse_x.toInt(),
					mouse_y.toInt(),
					font ?: screen.font,
					itemstack
				)
				// GuiUtils.postItemToolTip();

				return true
			}
		}
		return false
	}

	companion object {
		val SLOT_HIGHLIGHT = RGBAColor(255, 255, 255, 100)
		val SLOT_HIGHLIGHT_DRAG = RGBAColor(200, 200, 200, 150)
		const val SIZE = 18f
		val SLOT_BACKGROUND = WidgetLocation.WIDGETS.element(0f, 0f, SIZE, SIZE)
	}
}
