package ru.dbotthepony.mc.otm.client.screen.panels

import com.mojang.blaze3d.platform.InputConstants
import com.mojang.blaze3d.vertex.PoseStack
import ru.dbotthepony.mc.otm.client.screen.MatteryScreen
import kotlin.math.roundToInt

open class DiscreteScrollBarPanel(
	screen: MatteryScreen<*>,
	parent: EditablePanel?,
	val maxScroll: (panel: DiscreteScrollBarPanel) -> Int,
	val scrollCallback: (panel: DiscreteScrollBarPanel, oldScroll: Int, newScroll: Int) -> Unit,
	x: Float = 0f,
	y: Float = 0f,
	height: Float = 20f
) : EditablePanel(screen, parent, x, y, width = ScrollBarConstants.WIDTH, height = height) {
	open inner class Button : EditablePanel(screen, this@DiscreteScrollBarPanel, 1f, 1f, 12f, 15f) {
		var isScrolling = false
			protected set

		override fun innerRender(stack: PoseStack, mouseX: Float, mouseY: Float, partialTick: Float) {
			if (isScrolling) {
				ScrollBarConstants.scrollBarButtonPress.render(stack, width = width, height = height)
			} else if (maxScroll.invoke(this@DiscreteScrollBarPanel) <= 0) {
				ScrollBarConstants.scrollBarButtonDisabled.render(stack, width = width, height = height)
			} else if (isHovered) {
				ScrollBarConstants.scrollBarButtonHover.render(stack, width = width, height = height)
			} else {
				ScrollBarConstants.scrollBarButton.render(stack, width = width, height = height)
			}
		}

		protected var rememberScroll = 0
		private var rememberY = 0.0

		override fun mouseClickedInner(x: Double, y: Double, button: Int): Boolean {
			if (maxScroll.invoke(this@DiscreteScrollBarPanel) <= 0) {
				return true
			}

			if (button == InputConstants.MOUSE_BUTTON_LEFT) {
				isScrolling = true
				trapMouseInput = true
				rememberScroll = scroll
				rememberY = y
			}

			return true
		}

		override fun mouseReleasedInner(x: Double, y: Double, button: Int): Boolean {
			if (isScrolling && button == InputConstants.MOUSE_BUTTON_LEFT) {
				isScrolling = false
				trapMouseInput = false
				return true
			}

			return false
		}

		override fun mouseScrolledInner(x: Double, y: Double, scroll: Double): Boolean {
			return this@DiscreteScrollBarPanel.mouseScrolledInner(x, y, scroll)
		}

		override fun mouseDraggedInner(x: Double, y: Double, button: Int, xDelta: Double, yDelta: Double): Boolean {
			if (isScrolling) {
				val pixelsPerRow = (this@DiscreteScrollBarPanel.height - height) / maxScroll.invoke(this@DiscreteScrollBarPanel)
				val diff = y - rememberY

				this@DiscreteScrollBarPanel.scroll = rememberScroll + (diff / pixelsPerRow).roundToInt()
				return true
			}

			return false
		}
	}

	val scrollButton = Button()

	override fun innerRender(stack: PoseStack, mouseX: Float, mouseY: Float, partialTick: Float) {
		ScrollBarConstants.scrollBarBody.render(stack, y = 2f, height = height - 4)
		ScrollBarConstants.scrollBarTop.render(stack)
		ScrollBarConstants.scrollBarBottom.render(stack, y = height - 2)
	}

	public override fun mouseScrolledInner(x: Double, y: Double, scroll: Double): Boolean {
		this.scroll -= scroll.toInt()
		return true
	}

	var scroll = 0
		set(value) {
			if (value == field) {
				return
			}

			val newValue = if (value < 0) {
				0
			} else if (value > maxScroll.invoke(this)) {
				maxScroll.invoke(this)
			} else {
				value
			}

			if (newValue != field) {
				val old = field
				field = newValue
				scrollChanged(old, newValue)
			}
		}

	open fun updateScrollButtonPosition() {
		val maxScroll = this.maxScroll.invoke(this)

		if (maxScroll <= 0) {
			scrollButton.y = 1f
			return
		}

		val availableHeight = height - scrollButton.height - 2f
		scrollButton.y = 1f + availableHeight * (scroll.toFloat() / maxScroll)
	}

	open fun scrollChanged(oldScroll: Int, newScroll: Int) {
		scrollCallback.invoke(this, oldScroll, newScroll)
		updateScrollButtonPosition()
	}

	override fun performLayout() {
		super.performLayout()
		updateScrollButtonPosition()
	}
}
