package ru.dbotthepony.mc.otm.client.screen.widget

import com.mojang.blaze3d.vertex.PoseStack
import net.minecraft.network.chat.Component
import ru.dbotthepony.mc.otm.core.TranslatableComponent
import ru.dbotthepony.mc.otm.client.render.*
import ru.dbotthepony.mc.otm.client.screen.MatteryScreen
import ru.dbotthepony.mc.otm.client.screen.panels.EditablePanel
import ru.dbotthepony.mc.otm.core.formatPowerLevel
import ru.dbotthepony.mc.otm.menu.widget.LevelGaugeWidget

open class PowerGaugePanel(
	screen: MatteryScreen<*>,
	parent: EditablePanel? = null,
	val widget: LevelGaugeWidget,
	x: Float = 0f,
	y: Float = 0f,
	width: Float = GAUGE_BACKGROUND.w,
	height: Float = GAUGE_BACKGROUND.h
) : EditablePanel(screen, parent, x, y, width, height) {
	init {
		scissor = true
	}

	protected open fun makeTooltip(): MutableList<Component> {
		return mutableListOf(
			TranslatableComponent("otm.gui.power.percentage_level", String.format("%.2f", widget.percentage() * 100.0)),
			formatPowerLevel(widget.level(), widget.maxLevel())
		)
	}

	override fun innerRender(stack: PoseStack, mouseX: Float, mouseY: Float, partialTick: Float) {
		val height = this.height * widget.percentage()

		if (width >= 18f) {
			GAUGE_BACKGROUND_WIDE.render(stack, width = width, height = this.height)
			GAUGE_FOREGROUND_WIDE.renderPartial(
				stack,
				y = this.height - height,
				height = height,
				width = width,
				winding = UVWindingOrder.U0_V1_U1_V0)
		} else {
			GAUGE_BACKGROUND.render(stack, width = width, height = this.height)
			GAUGE_FOREGROUND.renderPartial(
				stack,
				y = this.height - height,
				height = height,
				width = width,
				winding = UVWindingOrder.U0_V1_U1_V0)
		}
	}

	override fun innerRenderTooltips(stack: PoseStack, mouse_x: Float, mouse_y: Float, flag: Float): Boolean {
		if (isHovered) {
			screen.renderComponentTooltip(stack, makeTooltip(), mouse_x.toInt(), mouse_y.toInt())
			return true
		}

		return false
	}

	companion object {
		val GAUGE_BACKGROUND = WidgetLocation.WIDGETS.element(x = 202f, y = 0f, w = 9f, h = 48f)
		val GAUGE_FOREGROUND = WidgetLocation.WIDGETS.element(x = 211f, y = 0f, w = 9f, h = 48f)

		val GAUGE_BACKGROUND_WIDE = WidgetLocation.WIDGETS.element(x = 238f, y = 0f, w = 18f, h = 48f)
		val GAUGE_FOREGROUND_WIDE = WidgetLocation.WIDGETS.element(x = 220f, y = 0f, w = 18f, h = 48f)
	}
}

/**
 * Shortcut to [PowerGaugePanel] with doubled width
 */
fun WidePowerGaugePanel(
	screen: MatteryScreen<*>,
	parent: EditablePanel? = null,
	widget: LevelGaugeWidget,
	x: Float = 0f,
	y: Float = 0f,
	width: Float = 18f,
	height: Float = 48f
) = PowerGaugePanel(screen, parent, widget, x, y, width, height)
