package ru.dbotthepony.mc.otm.menu

import net.minecraft.world.SimpleContainer
import net.minecraft.world.entity.player.Inventory
import net.minecraft.world.inventory.Slot
import net.minecraft.world.item.ItemStack
import ru.dbotthepony.mc.otm.core.ImmutableList
import ru.dbotthepony.mc.otm.block.entity.matter.MatterBottlerBlockEntity
import ru.dbotthepony.mc.otm.capability.MatteryCapability
import ru.dbotthepony.mc.otm.menu.widget.BooleanPlayerInputWidget
import ru.dbotthepony.mc.otm.menu.widget.LevelGaugeWidget
import ru.dbotthepony.mc.otm.menu.widget.ProgressGaugeWidget
import ru.dbotthepony.mc.otm.core.orNull
import ru.dbotthepony.mc.otm.registry.MMenus

class MatterBottlerMenu @JvmOverloads constructor(
	p_38852_: Int,
	inventory: Inventory,
	tile: MatterBottlerBlockEntity? = null
) : MatteryPoweredMenu(
	MMenus.MATTER_BOTTLER, p_38852_, inventory, tile
) {
	val workFlow: BooleanPlayerInputWidget

	val progressWidget: ProgressGaugeWidget
	val matterWidget: LevelGaugeWidget

	override val storageSlots: List<MatterySlot>

	init {
		val container = tile?.container ?: SimpleContainer(6)

		if (tile == null) {
			progressWidget = ProgressGaugeWidget(this)
			matterWidget = LevelGaugeWidget(this)
			workFlow = BooleanPlayerInputWidget(this).asClient()
		} else {
			progressWidget = ProgressGaugeWidget(this) { tile.getWorkProgress() }
			matterWidget = LevelGaugeWidget(this, tile.matter)
			workFlow = BooleanPlayerInputWidget(this, tile::workFlow)
		}

		storageSlots = ImmutableList(6) { index ->
			object : MatterySlot(container, index) {
				override fun mayPlace(p_40231_: ItemStack): Boolean {
					val cap = p_40231_.getCapability(MatteryCapability.MATTER).orNull() ?: return false

					if (workFlow.value) {
						return index < 3 && cap.allowsReceive
					} else {
						return index >= 3 && cap.allowsExtract
					}
				}
			}
		}

		storageSlots.forEach(this::addSlot)
		addInventorySlots()
	}

	override fun broadcastChanges() {
		super.broadcastChanges()
		workFlow.value = (tile as MatterBottlerBlockEntity).workFlow
	}
}
