
@file:Suppress("unused")

package ru.dbotthepony.mc.otm.registry

import net.minecraft.resources.ResourceLocation
import net.minecraftforge.eventbus.api.IEventBus
import net.minecraftforge.fml.javafmlmod.FMLJavaModLoadingContext
import net.minecraftforge.registries.DeferredRegister
import net.minecraftforge.registries.RegistryObject
import ru.dbotthepony.mc.otm.OverdriveThatMatters
import ru.dbotthepony.mc.otm.core.TextComponent
import ru.dbotthepony.mc.otm.core.TranslatableComponent
import ru.dbotthepony.mc.otm.android.AndroidResearchBuilder
import ru.dbotthepony.mc.otm.android.AndroidResearchType
import ru.dbotthepony.mc.otm.android.feature.NanobotsArmor
import ru.dbotthepony.mc.otm.client.render.SkinElement
import ru.dbotthepony.mc.otm.client.render.SkinGrid

object AndroidResearch {
	val ICONS = ResourceLocation(OverdriveThatMatters.MOD_ID, "textures/gui/android_upgrades.png")
	val ICON_TRANSFER: SkinElement
	val ICON_ATTACK_BOOST: SkinElement
	val ICON_PLASMA_SHIELD_BOOST: SkinElement
	val ICON_CLOAK: SkinElement
	val ICON_GRAVITATIONAL_STABILIZER: SkinElement
	val ICON_AIR_BAGS: SkinElement
	val ICON_JUMP_BOOST: SkinElement

	val ICON_FEATHER_FALLING: SkinElement
	val ICON_ARC: SkinElement
	val ICON_ARROW: SkinElement
	val ICON_ARMOR: SkinElement
	val ICON_NANOBOTS: SkinElement
	val ICON_NIGHT_VISION: SkinElement
	val ICON_OXYGEN_SUPPLY: SkinElement

	val ICON_PLASMA_SHIELD: SkinElement
	val ICON_SHOCKWAVE: SkinElement
	val ICON_LIMB_OVERCLOCKING: SkinElement
	val ICON_STEP_ASSIST: SkinElement
	val ICON_ENDER_TELEPORT: SkinElement
	val ICON_WIRELESS_CHARGING: SkinElement
	val ICON_UNKNOWN: SkinElement

	val ICON_EXTENDED_REACH: SkinElement

	init {
		val grid = SkinGrid(ICONS, 18f, 18f, 7, 7)

		ICON_TRANSFER = grid.next()
		ICON_ATTACK_BOOST = grid.next()
		ICON_PLASMA_SHIELD_BOOST = grid.next()
		ICON_CLOAK = grid.next()
		ICON_GRAVITATIONAL_STABILIZER = grid.next()
		ICON_AIR_BAGS = grid.next()
		ICON_JUMP_BOOST = grid.next()

		ICON_FEATHER_FALLING = grid.next()
		ICON_ARC = grid.next()
		ICON_ARROW = grid.next()
		ICON_ARMOR = grid.next()
		ICON_NANOBOTS = grid.next()
		ICON_NIGHT_VISION = grid.next()
		ICON_OXYGEN_SUPPLY = grid.next()

		ICON_PLASMA_SHIELD = grid.next()
		ICON_SHOCKWAVE = grid.next()
		ICON_LIMB_OVERCLOCKING = grid.next()
		ICON_STEP_ASSIST = grid.next()
		ICON_ENDER_TELEPORT = grid.next()
		ICON_WIRELESS_CHARGING = grid.next()
		ICON_UNKNOWN = grid.next()

		ICON_EXTENDED_REACH = grid.next()
	}

	private val registry = DeferredRegister.create(MRegistry.ANDROID_RESEARCH_KEY, OverdriveThatMatters.MOD_ID)

	internal fun register(bus: IEventBus) {
		registry.register(bus)
	}

	val AIR_BAGS: AndroidResearchType<*> by registry.register(MNames.AIR_BAGS) {
		AndroidResearchBuilder()
			.withExperience(18)
			.addFeatureResult(AndroidFeatures.AIR_BAGS)
			.withDescription()
			.withIcon(ICON_AIR_BAGS)
			.build()
	}

	val IMPROVED_LIMBS: AndroidResearchType<*> by registry.register(MNames.IMPROVED_LIMBS) {
		AndroidResearchBuilder()
			.withExperience(20)
			.withDescription()
			.withIcon(ICON_EXTENDED_REACH)
			.build()
	}

	val STEP_ASSIST: AndroidResearchType<*> by registry.register(MNames.STEP_ASSIST) {
		AndroidResearchBuilder()
			.withExperience(24)
			.addFeatureResult(AndroidFeatures.STEP_ASSIST)
			.withDescription()
			.withIcon(ICON_STEP_ASSIST)
			.addPrerequisite(IMPROVED_LIMBS)
			.build()
	}

	val EXTENDED_REACH: AndroidResearchType<*> by registry.register(MNames.EXTENDED_REACH) {
		AndroidResearchBuilder()
			.withExperience(40)
			.addFeatureResult(AndroidFeatures.EXTENDED_REACH)
			.addPrerequisite(IMPROVED_LIMBS)
			.withDescription()
			.withIcon(ICON_EXTENDED_REACH)
			.build()
	}

	val NANOBOTS: AndroidResearchType<*> by registry.register(MNames.NANOBOTS) {
		AndroidResearchBuilder()
			.withExperience(15)
			.withDescription()
			.withIcon(ICON_NANOBOTS)
			.build()
	}

	val NANOBOTS_ARMOR: AndroidResearchType<*> by registry.register(MNames.NANOBOTS_ARMOR) {
		AndroidResearchBuilder()
			.withExperience(25)
			.withDescription()
			.addPrerequisite(OverdriveThatMatters.loc(MNames.NANOBOTS))
			.addFeatureResult(OverdriveThatMatters.loc(MNames.NANOBOTS_ARMOR))
			.withIcon(ICON_ARMOR)
			.addBlocker(OverdriveThatMatters.loc(MNames.ATTACK_BOOST_1))
			.build()
	}

	val LIMB_OVERCLOCKING: List<AndroidResearchType<*>>
	val NANOBOTS_ARMOR_SPEED: List<AndroidResearchType<*>>
	val NANOBOTS_ARMOR_STRENGTH: List<AndroidResearchType<*>>
	val NANOBOTS_REGENERATION: List<AndroidResearchType<*>>
	val ATTACK_BOOST: List<AndroidResearchType<*>>

	init {
		val limbList = ArrayList<RegistryObject<AndroidResearchType<*>>>()
		val regenList = ArrayList<RegistryObject<AndroidResearchType<*>>>()
		val armorStrengthList = ArrayList<RegistryObject<AndroidResearchType<*>>>()
		val armorSpeedList = ArrayList<RegistryObject<AndroidResearchType<*>>>()
		val attackBoostList = ArrayList<RegistryObject<AndroidResearchType<*>>>()

		for (i in 0 until 4) {
			limbList.add(registry.register(MNames.LIMB_OVERCLOCKING_LIST[i]) {
				val research = AndroidResearchBuilder()
					.withExperience(18 + i * 8)
					.withIconText(TextComponent((i + 1).toString()))
					.addPrerequisite(IMPROVED_LIMBS)
					.withIcon(ICON_LIMB_OVERCLOCKING)
					.withName(TranslatableComponent("android_research.overdrive_that_matters.limb_overclocking", i + 1))
					.withDescription(
						TranslatableComponent(
							"android_research.overdrive_that_matters.limb_overclocking.description",
							(i + 1) * 8,
							(i + 1) * 6
						)
					)
					.addFeatureResult(OverdriveThatMatters.loc(MNames.LIMB_OVERCLOCKING), i)

				if (i > 0) {
					research.addPrerequisite(OverdriveThatMatters.loc(MNames.LIMB_OVERCLOCKING_LIST[i - 1]))
				}

				research.build()
			})

			attackBoostList.add(registry.register(MNames.ATTACK_BOOST_LIST[i]) {
				val research = AndroidResearchBuilder()
					.withExperience(18 + i * 8)
					.withIconText(TextComponent((i + 1).toString()))
					.addPrerequisite(IMPROVED_LIMBS)
					.withIcon(ICON_ATTACK_BOOST)
					.withName(TranslatableComponent("android_research.overdrive_that_matters.attack_boost", i + 1))
					.withDescription(
						TranslatableComponent(
							"android_research.overdrive_that_matters.attack_boost.description",
							(i + 1) * 6
						)
					)
					.addFeatureResult(OverdriveThatMatters.loc(MNames.ATTACK_BOOST), i)
					.addBlocker(NANOBOTS_ARMOR)

				if (i > 0) {
					research.addPrerequisite(OverdriveThatMatters.loc(MNames.ATTACK_BOOST_LIST[i - 1]))
				}

				research.build()
			})

			regenList.add(registry.register(MNames.NANOBOTS_REGENERATION_LIST[i]) {
				val regeneration = AndroidResearchBuilder()
					.withExperience(20 + i * 6)
					.withIconText(TextComponent((i + 1).toString()))
					.withIcon(ICON_NANOBOTS)
					.withName(TranslatableComponent("android_research.overdrive_that_matters.nanobots_regeneration", i + 1))
					.withDescription(
						if (i > 0) TranslatableComponent("android_research.overdrive_that_matters.nanobots_regeneration.description_improve") else TranslatableComponent(
							"android_research.overdrive_that_matters.nanobots_regeneration.description"
						)
					)
					.addFeatureResult(OverdriveThatMatters.loc(MNames.NANOBOTS_REGENERATION), i)


				if (i > 0) {
					regeneration.addPrerequisite(OverdriveThatMatters.loc(MNames.NANOBOTS_REGENERATION_LIST[i - 1]))
				} else {
					regeneration.addPrerequisite(OverdriveThatMatters.loc(MNames.NANOBOTS))
				}

				regeneration.build()
			})
		}

		for (i in 0 until 3) {
			val level = i + 1

			armorStrengthList.add(registry.register(MNames.NANOBOTS_ARMOR_STRENGTH_LIST[i]) {
				AndroidResearchBuilder()
					.withExperience(20 + i * 8)
					.withIconText(TextComponent((i + 1).toString()))
					.withIcon(ICON_ARMOR)
					.addPrerequisite(OverdriveThatMatters.loc(if (i > 0) MNames.NANOBOTS_ARMOR_STRENGTH_LIST[i - 1] else MNames.NANOBOTS_ARMOR))
					.withName(
						TranslatableComponent(
							"android_research.overdrive_that_matters.nanobots_armor_strength",
							i + 1
						)
					)
					.withDescription(
						TranslatableComponent(
							"android_research.overdrive_that_matters.nanobots_armor_strength.description",
							(i + 1) * 8,
							(i + 1) * 6
						)
					)
					.addFeatureResult(OverdriveThatMatters.loc(MNames.NANOBOTS_ARMOR), 0) { _, feature ->
						if ((feature as NanobotsArmor).strength < level) feature.strength = level
					}
					.build()
			})

			armorSpeedList.add(registry.register(MNames.NANOBOTS_ARMOR_SPEED_LIST[i]) {
				AndroidResearchBuilder()
					.withExperience(20 + i * 8)
					.withIconText(TextComponent((i + 1).toString()))
					.withIcon(ICON_ARMOR)
					.addPrerequisite(OverdriveThatMatters.loc(if (i > 0) MNames.NANOBOTS_ARMOR_SPEED_LIST[i - 1] else MNames.NANOBOTS_ARMOR))
					.withName(
						TranslatableComponent(
							"android_research.overdrive_that_matters.nanobots_armor_speed",
							i + 1
						)
					)
					.withDescription(
						TranslatableComponent(
							"android_research.overdrive_that_matters.nanobots_armor_speed.description",
							(i + 1) * 8,
							(i + 1) * 6
						)
					)
					.addFeatureResult(OverdriveThatMatters.loc(MNames.NANOBOTS_ARMOR), 0) { _, feature ->
						if ((feature as NanobotsArmor).speed < level) feature.speed = level
					}
					.build()
			})
		}

		LIMB_OVERCLOCKING = RegistryObjectList(limbList)
		NANOBOTS_REGENERATION = RegistryObjectList(regenList)
		NANOBOTS_ARMOR_SPEED = RegistryObjectList(armorSpeedList)
		NANOBOTS_ARMOR_STRENGTH = RegistryObjectList(armorStrengthList)
		ATTACK_BOOST = RegistryObjectList(attackBoostList)
	}
}
