package ru.dbotthepony.mc.otm.registry

import com.google.common.collect.ImmutableList
import com.google.common.collect.ImmutableSet
import net.minecraft.world.item.DyeColor
import net.minecraftforge.registries.RegistryObject

class LazyList<T>(private val getters: ImmutableList<() -> T>) : AbstractList<T>() {
	constructor(vararg getters: () -> T) : this(ImmutableList.copyOf(getters))
	constructor(getters: List<() -> T>) : this(ImmutableList.copyOf(getters))

	override val size: Int
		get() = getters.size

	override fun get(index: Int): T {
		return getters[index].invoke()
	}
}

class LazyMap<K, T> : AbstractMap<K, T> {
	override val entries: Set<Map.Entry<K, T>>

	constructor(vararg mValues: Pair<K, () -> T>) : super() {
		entries = ImmutableSet.copyOf(mValues.map { Entry(it.first, it.second) })
	}

	constructor(mValues: List<Pair<K, () -> T>>) : super() {
		entries = ImmutableSet.copyOf(mValues.map { Entry(it.first, it.second) })
	}

	constructor(mValues: Map<K, () -> T>) : super() {
		entries = ImmutableSet.copyOf(mValues.map { Entry(it.key, it.value) })
	}

	private inner class Entry(
		override val key: K,
		private val getter: () -> T
	) : Map.Entry<K, T> {
		override val value: T
			get() = getter.invoke()
	}
}

class RegistryObjectList<T>(private vararg val getters: RegistryObject<T>) : AbstractList<T>() {
	constructor(getters: List<RegistryObject<T>>) : this(*getters.toTypedArray())

	override val size: Int
		get() = getters.size

	override fun get(index: Int): T {
		return getters[index].get()
	}
}
